from abc import ABCMeta, abstractmethod
import csv
import enum
from flask import abort, request, Response
from flask.views import View
from io import BytesIO
from odf.style import TextProperties, TableRowProperties, TableColumnProperties, TableCellProperties, FontFace

from dmutils import csv_generator
from dmutils import ods


class DownloadFileView(View, metaclass=ABCMeta):
    """An abstract base class appropriate for subclassing in the frontend apps when the user needs to be able to
    download some data as a CSV or ODS file. All abstract methods must be implemented on the subclass (although example
    implementations are included here with the kind of return value expected); all other methods should be able to
    be left alone to support handling and dispatching the request."""

    FILETYPES = enum.Enum('Filetypes', ['CSV', 'ODS'])

    def __init__(self, **kwargs):
        self.request = request

        self.data_api_client = None
        self.search_api_client = None
        self.content_loader = None

        super(View, self).__init__(**kwargs)

    @abstractmethod
    def _init_hook(self, **kwargs):
        """Hook into the start of the request to associate required clients/loaders with the instance and perform any
        other logic required for the start of the request."""
        self.data_api_client = None
        self.search_api_client = None
        self.content_loader = None

    def _post_request_hook(self, response=None, file_context=None, **kwargs):
        """A hook that can be used to implement any logic that should be run after the main download file process,
        and immediately before sending the response to the user."""
        pass

    @abstractmethod
    def get_file_context(self, **kwargs):
        """Must return a dictionary containing any data required by the generation routines.

        Required keys:
         * `filename`: without a file extension
         * `sheetname`: the name of the first sheet in the ODS file (assumption: single sheeted spreadsheet)"""
        return {'filename': 'my-download', 'sheetname': 'Sheet 1'}

    @abstractmethod
    def determine_filetype(self, file_context=None, **kwargs):
        """Logic to tell the View which filetype to generate; must return a choice from DownloadFileView.FILETYPES"""
        return DownloadFileView.FILETYPES.ODS

    @abstractmethod
    def generate_csv_rows(self, file_context):
        """Must return a nested iterable of rows+cells, eg [[heading1, heading2], [data1, data2], ...]"""
        return [['Heading 1', 'Heading 2'], ['Row 1, Column 1', 'Row 1, Column 2']]

    @abstractmethod
    def populate_styled_ods_with_data(self, spreadsheet, file_context):
        """Takes an empty dmutils.ods.SpreadSheet (with required fonts/styles already added) and populates the required
        data into it."""
        sheet = spreadsheet.sheet(file_context['sheetname'])
        sheet.write_row(name='header', cells=['Heading 1', 'Heading 2'])
        sheet.write_row(name='row1', cells=['Row 1, Column 1', 'Row 1, Column 2'])

    @staticmethod
    def create_blank_ods_with_styles():
        """Create a dmutils.ods.SpreadSheet pre-configured with some default styles, ready for population with data
        appropriate for the subclass View. Modifications here (except adding styles) are likely breaking changes."""
        spreadsheet = ods.SpreadSheet()

        # Add the font we will use for the entire spreadsheet.
        spreadsheet.add_font(FontFace(name="Arial", fontfamily="Arial"))

        # Add some default styles for columns.
        spreadsheet.add_style("col-default", "table-column", (
            TableColumnProperties(breakbefore="auto"),
        ), parentstylename="Default")

        spreadsheet.add_style("col-wide", "table-column", (
            TableColumnProperties(columnwidth="150pt", breakbefore="auto"),
        ), parentstylename="Default")

        spreadsheet.add_style("col-extra-wide", "table-column", (
            TableColumnProperties(columnwidth="300pt", breakbefore="auto"),
        ), parentstylename="Default")

        # Add some default styles for rows.
        spreadsheet.add_style("row-default", "table-row", (
            TableRowProperties(breakbefore="auto", useoptimalrowheight="false"),
        ), parentstylename="Default")

        spreadsheet.add_style("row-tall", "table-row", (
            TableRowProperties(breakbefore="auto", rowheight="30pt", useoptimalrowheight="false"),
        ), parentstylename="Default")

        spreadsheet.add_style("row-tall-optimal", "table-row", (
            TableRowProperties(breakbefore="auto", rowheight="30pt", useoptimalrowheight="true"),
        ), parentstylename="Default")

        # Add some default styles for cells.
        spreadsheet.add_style("cell-default", "table-cell", (
            TableCellProperties(wrapoption="wrap", verticalalign="top"),
            TextProperties(fontfamily="Arial", fontnameasian="Arial", fontnamecomplex="Arial", fontsize="11pt"),
        ), parentstylename="Default")

        spreadsheet.add_style("cell-header", "table-cell", (
            TableCellProperties(wrapoption="wrap", verticalalign="top"),
            TextProperties(fontfamily="Arial", fontnameasian="Arial", fontnamecomplex="Arial", fontsize="11pt",
                           fontweight="bold"),
        ), parentstylename="Default")

        return spreadsheet

    def create_response(self, file_context, file_type):
        if file_type == DownloadFileView.FILETYPES.CSV:
            body = csv_generator.iter_csv(self.generate_csv_rows(file_context), quoting=csv.QUOTE_ALL)

            mimetype = 'text/csv; header=present'

        elif file_type == DownloadFileView.FILETYPES.ODS:
            buffer = BytesIO()

            ods = self.create_blank_ods_with_styles()
            self.populate_styled_ods_with_data(ods, file_context)
            ods.save(buffer)

            body = buffer.getvalue()

            mimetype = 'application/vnd.oasis.opendocument.spreadsheet'

        else:
            abort(400)

        content_disposition = 'attachment;filename={}.{}'.format(file_context['filename'], file_type.name.lower())

        return Response(
            body,
            mimetype=mimetype,
            headers={
                "Content-Disposition": content_disposition,
                "Content-Type": mimetype
            }
        ), 200

    def dispatch_request(self, **kwargs):
        self._init_hook(**kwargs)

        file_context = self.get_file_context(**kwargs)
        file_type = self.determine_filetype(file_context, **kwargs)

        if not isinstance(file_type, enum.Enum) or file_type.name not in DownloadFileView.FILETYPES.__members__:
            abort(400)

        response = self.create_response(file_context, file_type)

        self._post_request_hook(response, **kwargs)
        return response


class SimpleDownloadFileView(DownloadFileView, metaclass=ABCMeta):
    """A slightly simplier version of the DownloadFileView where it is possible to have a single source of data
    for all download filetypes. If you want a fairly simple structure to your spreadsheet (in effect, a single sheet,
    with self-contained rows where cells don't span rows or columns), you can implement only `get_file_data_and_styles`
    to work with all supported filetypes."""
    @abstractmethod
    def get_file_data_and_column_styles(self, file_context):
        """Example implementation with basic styling"""
        file_rows = []

        # Assign the column styles
        column_styles = [
            {'stylename': 'col-default', 'defaultcellstylename': 'cell-default'},  # Header 1
            {'stylename': 'col-default', 'defaultcellstylename': 'cell-default'},  # Header 2
        ]

        # Headers
        file_rows.append({
            'cells': ['Header 1', 'Header 2'],
            'meta': {'name': 'header',
                     'row_styles': {'stylename': 'row-default'},
                     'cell_styles': {'stylename': 'cell-header'}},
        })

        # Data
        rows = [['Row 1, Column 1', 'Row 1, Column 2'], ['Row 2, Column 1', 'Row 2, Column 2']]
        for i, row in enumerate(rows):
            file_rows.append({
                'cells': row,
                'meta': {'name': 'row-{}'.format(i),
                         'row_styles': {'stylename': 'row-default'},
                         'cell_styles': {'stylename': 'cell-default'}},
            })

        return file_rows, column_styles

    def generate_csv_rows(self, file_context):
        """Must return a nested iterable of iterables containing the data to be written out to csv"""
        file_rows, _ = self.get_file_data_and_column_styles(file_context)

        return [row['cells'] for row in file_rows]

    def populate_styled_ods_with_data(self, spreadsheet, file_context):
        """Must return an instance of dmutils.ods.SpreadSheet populated with the required data."""
        file_rows, column_styles = self.get_file_data_and_column_styles(file_context)
        sheet = spreadsheet.sheet(file_context.get('sheetname', 'Sheet 1'))

        # Add the column styles
        for column_style in column_styles:
            sheet.create_column(**column_style)

        # Add the data
        for file_row in file_rows:
            row_name = file_row['meta'].get('name')
            write_row_kwargs = file_row['meta'].copy()
            del write_row_kwargs['name']

            sheet.write_row(name=row_name, cells=file_row['cells'], **write_row_kwargs)



"""
Common utils for Digital Marketplace apps.
"""
import re
import ast
from setuptools import setup, find_packages


_version_re = re.compile(r'__version__\s+=\s+(.*)')

with open('dmutils/__init__.py', 'rb') as f:
    version = str(ast.literal_eval(_version_re.search(
        f.read().decode('utf-8')).group(1)))

setup(
    name='digitalmarketplace-utils',
    version=version,
    url='https://github.com/alphagov/digitalmarketplace-utils',
    license='MIT',
    author='GDS Developers',
    description='Common utils for Digital Marketplace apps.',
    long_description=__doc__,
    packages=find_packages(),
    package_data={'dmutils': ['py.typed']},
    include_package_data=True,
    install_requires=[
         'Flask-WTF>=0.14.2',
         'Flask>=1.0,<2.1',
         'Flask-gzip>=0.2',
         'Flask-Login>=0.2.11',
         'Flask-Session>=0.3.2',
         'boto3<2,>=1.7.83',
         'contextlib2>=0.4.0',
         'cryptography>=3.2',
         'gds-metrics>=0.2.0,<1',
         'govuk-country-register>=0.3.0',
         'mailchimp3==3.0.15',
         'requests>=2.22.0,<3',
         'redis>=3.5.2',
         'fleep<1.1,>=1.0.1',
         'notifications-python-client>=5.0.1,<7.0.0',
         'odfpy>=1.3.6',
         'python-json-logger>=0.1.11,<3.0.0',
         'pytz',
         'unicodecsv>=0.14.1',
         'workdays>=1.4',
    ],
    python_requires="~=3.8",
)



from functools import wraps
from datetime import datetime, timedelta

from flask import current_app


def deprecated(dies_at):
    """Mark a flask view as deprecated

    Usage:
    @deprecated(dies_at=datetime(2015, 8, 1))
    """
    def decorator(view):
        @wraps(view)
        def func(*args, **kwargs):
            message = "Calling deprecated view '{view_name}'. Dies in {time_left}."
            time_left = dies_at - datetime.utcnow()
            extra = {'view_name': view.__name__, 'time_left': time_left}
            if time_left < timedelta(days=7):
                current_app.logger.error(message, extra=extra)
            else:
                current_app.logger.warning(message, extra=extra)
            response = view(*args, **kwargs)
            response.headers['DM-Deprecated'] = "Dies in {}".format(time_left)
            return response

        return func

    return decorator



import os


def init_app(app):
    for key, value in app.config.items():
        if key in os.environ:
            if isinstance(value, bool):
                app.config[key] = _convert_to_boolean_or_fail(key, os.environ[key])
            elif isinstance(value, int):
                app.config[key] = _convert_to_int_or_fail(key, os.environ[key])
            else:
                app.config[key] = os.environ[key]
    app.config['DM_ENVIRONMENT'] = os.environ.get('DM_ENVIRONMENT',
                                                  'development')

    # From Flask 1.0 onwards, ENV is set to either 'production' (default) or 'development'.
    # If set as 'development' then debug mode will be enabled.
    # If left as 'production' when running a local development server, it gives an unsettling warning message.
    # See http://flask.pocoo.org/docs/1.0/config/#environment-and-debug-features
    if app.config['DM_ENVIRONMENT'] == 'development':
        app.config['ENV'] = 'development'


def _convert_to_boolean_or_fail(key, value):
    result = convert_to_boolean(value)
    if not isinstance(result, bool):
        raise ValueError("{} must be boolean".format(key))
    return result


def convert_to_boolean(value):
    """Turn strings to bools if they look like them

    Truthy things should be True
    >>> for truthy in ['true', 'on', 'yes', '1']:
    ...   assert convert_to_boolean(truthy) == True

    Falsey things should be False
    >>> for falsey in ['false', 'off', 'no', '0']:
    ...   assert convert_to_boolean(falsey) == False

    Other things should be unchanged
    >>> for value in ['falsey', 'other', True, 0]:
    ...   assert convert_to_boolean(value) == value
    """
    if isinstance(value, str):
        if value.lower() in ['t', 'true', 'on', 'yes', '1']:
            return True
        elif value.lower() in ['f', 'false', 'off', 'no', '0']:
            return False

    return value


def _convert_to_int_or_fail(key, value):
    result = convert_to_number(value)
    if not isinstance(result, int):
        raise ValueError("{} must be an integer".format(key))
    return result


def convert_to_number(value):
    """Turns numeric looking things into floats or ints

    Integery things should be integers
    >>> for inty in ['0', '1', '2', '99999']:
    ...   assert isinstance(convert_to_number(inty), int)

    Floaty things should be floats
    >>> for floaty in ['0.99', '1.1', '1000.0000001']:
    ...   assert isinstance(convert_to_number(floaty), float)

    Other things should be unchanged
    >>> for value in [0, 'other', True, 123]:
    ...   assert convert_to_number(value) == value
    """
    try:
        return float(value) if "." in value else int(value)
    except (TypeError, ValueError):
        return value



from collections import OrderedDict
import typing


# TODO use TypedDict once we're on python 3.8 to allow us to be more specific about this mapping's contents
ErrorMapping = typing.Mapping[str, typing.Any]


def govuk_error(error: ErrorMapping) -> ErrorMapping:
    if error:
        text = error.get("message", error.get("question", ""))
        return {
            "text": text,
            "href": error.get("href") or f"#input-{error['input_name']}",
            "errorMessage": {"text": text},
        }
    else:
        return {}


# TODO: use typing.OrderedDict once we're on Python 3.7
def govuk_errors(errors: typing.Mapping[str, ErrorMapping]) -> typing.Mapping[str, ErrorMapping]:
    """Converts digitalmarketplace-frontend-toolkit style errors into a format
    suitable for use with either digitalmarketplace-frontend-toolkit
    templates/macros or govuk-frontend macros.

    Returns a dictionary with the following keys, supporting three types of
    error display:

    1) digitalmarketplace-frontend-toolkit template toolkit/forms/validation.html:
      `input_name`, `question`, `message`

    2) govuk-frontend macro govukErrorSummary:
      `text`, `href`

    3) govuk-frontend errorMessage:
      `errorMessage`

    This allows us to treat errors from both content-loader forms and wtforms
    the same way inside templates,
    and provides backwards-compatible support for the migration to using
    govuk-frontend components.

    Example usage:

        # app.py

        @flask.route("/", methods=["GET", "POST"])
        def view():
            section = content_loader.get_section()
            try:
                data_api_client.update_something()
            except HTTPError as e:
                errors = govuk_errors(
                    section.get_error_messages(e.message)
                )

            return render(
                "template.html",
                question=section,
                errors=errors,
            )

        # template.html  (govuk-frontend)
        {{ govukErrorSummary({
            "errorList": errors.values()
        }) }}

        {{ govukTextInput({
            "errorMessage": errors.input.errorMessage,
        }) }}

        # template.html (DM frontend toolkit uses 'errors' by default)
        {% include 'toolkit/forms/validation.html' %}

        {{ forms[question.type](question, service_data, errors) }}

    :param errors: A digitalmarketplace-frontend-toolkit error dictionary
    :return: A dict with error information in a form suitable for
             Digital Marketplace templates
    """
    return OrderedDict(
        (key, {**error, **govuk_error(error)}) for key, error in errors.items()
    )


def get_errors_from_wtform(form):
    """Converts Flask-WTForm errors into formats suitable for use in Digital Marketplace templates.

    Returns a dictionary with the following keys, supporting three types of error display:

    1) digitalmarketplace-frontend-toolkit template toolkit/forms/validation.html:
      `input_name`, `question`, `message`

    2) govuk-frontend macro govukErrorSummary:
      `text`, `href`

    3) govuk-frontend errorMessage:
      `errorMessage`

    This allows us to treat errors from both content-loader forms and wtforms the same way inside templates,
    and provides backwards-compatible support for the migration to using govuk-frontend components.

    The `href` for the govuk-frontend error summary is derived from the `id` of the form field; if you need to
    override the `href` for some reason you can do this by passing the correct `id` to the field constructor.

    Example usage:

        # app.py
        class Form(FlaskForm):
            input = DMTextInput()

        @flask.route("/")
        def view():
            form = Form()
            errors = get_errors_from_wtform(form)
            return render(
                "template.html",
                errors=errors,
            )

        # template.html  (govuk-frontend)
        {{ govukErrorSummary({
            "errorList": errors.values()
        }) }}

        {{ govukTextInput({
            "errorMessage": errors.input.errorMessage,
        }) }}

        # template.html (DM frontend toolkit uses 'errors' by default)
        {% include 'toolkit/forms/validation.html' %}

    :param form: A Flask-WTForm
    :return: A dict with error information in a form suitable for Digital Marketplace templates
    """
    return OrderedDict(
        # TODO: remove legacy code (items for frontend toolkit validation banners)
        (
            key,
            {
                # parameters for digitalmarketplace-frontend-toolkit template toolkit/forms/validation.html
                "input_name": key, "question": form[key].label.text, "message": form[key].errors[0],

                # parameters for govuk-frontend macro govukErrorSummary
                "text": form[key].errors[0], "href": f"#{getattr(form[key], 'href', form[key].id)}",

                # parameters for govuk-frontend errorMessage parameter
                "errorMessage": (
                    {"text": form[key].errors[0]}
                    if form[key].errors[0]
                    else {}
                )
            }
        )
        for key in
        form.errors.keys()
    )



# -*- coding: utf-8 -*-
"""Digital Marketplace Notify integration."""

from typing import Dict, NamedTuple, Optional
import logging

from flask import current_app
from notifications_python_client import NotificationsAPIClient
from notifications_python_client.errors import HTTPError

from dmutils.email.exceptions import EmailError, EmailTemplateError, EmailInvalidError
from dmutils.email.helpers import hash_string
from dmutils.timing import logged_duration_for_external_request as log_external_request


class DMNotifyEmail(NamedTuple):
    to_email_address: str
    template_name_or_id: str
    reference: str
    personalisation: Optional[Dict[str, str]] = None


class DMNotifyClient:
    """Digital Marketplace wrapper around the Notify python client."""

    _client_class = NotificationsAPIClient
    _sent_references_cache = None

    def __init__(
            self,
            govuk_notify_api_key=None,
            govuk_notify_base_url='https://api.notifications.service.gov.uk',
            redirect_domains_to_address=None,
            *,
            logger=None,
            templates=None,
    ):
        """
            :param govuk_notify_api_key: defaults to current_app.config["DM_NOTIFY_API_KEY"]
            :param govuk_notify_base_url:
            :param redirect_domains_to_address: dictionary mapping email domain to redirected email address - emails
                sent to a email with a domain in this mapping will instead be sent to the corresponding value set here.
                if `redirect_domains_to_address` is `None` will fall back to looking for a
                `DM_NOTIFY_REDIRECT_DOMAINS_TO_ADDRESS` setting in the current flask app's config (if available).

        The following arguments are keyword-only and should only be used if you want to operate outside
        of a Flask app context.
            :param logger: logger to log progress to, taken from current_app if Falsey
            :param templates: a dictionary of template names to template uuids, so that you can use
                descriptive names when specifying a template. This defaults to current_app.config["NOTIFY_TEMPLATES"].
        """
        if govuk_notify_api_key is None:
            govuk_notify_api_key = current_app.config["DM_NOTIFY_API_KEY"]

        if templates is None:
            self.templates = current_app.config.get("NOTIFY_TEMPLATES", {}) if current_app else {}

        self.logger = logger or current_app.logger

        self.client = self._client_class(govuk_notify_api_key, govuk_notify_base_url)
        self._redirect_domains_to_address = (
            current_app.config.get("DM_NOTIFY_REDIRECT_DOMAINS_TO_ADDRESS")
            if current_app and redirect_domains_to_address is None else
            redirect_domains_to_address
        )

    def get_all_notifications(self, **kwargs):
        """Wrapper for notifications_python_client.notifications.NotificationsAPIClient::get_all_notifications"""
        with log_external_request(service='Notify'):
            return self.client.get_all_notifications(**kwargs)['notifications']

    def get_delivered_notifications(self, **kwargs):
        """Wrapper for notifications_python_client.notifications.NotificationsAPIClient::get_all_notifications"""
        return self.get_all_notifications(status='delivered', **kwargs)

    def get_delivered_references(self, invalidate_cache=False):
        """Get the references of all notifications that have already been delivered."""
        if invalidate_cache:
            self._sent_references_cache = None
        if self._sent_references_cache is None:
            self._sent_references_cache = set(i['reference'] for i in self.get_delivered_notifications())
        return self._sent_references_cache

    def _update_cache(self, reference):
        """If the cache has been instantiated then cache the new reference."""
        if self._sent_references_cache is not None:
            self._sent_references_cache.update([reference])

    def has_been_sent(self, reference, use_recent_cache=True):
        """
        Checks for a matching reference in our list of recently delivered references (last 250 emails).
        If use_recent_cache is set to False, we do a fresh lookup of the reference in the Notify API.
        """
        if not use_recent_cache:
            return len(self.client.get_all_notifications(reference=reference)['notifications']) > 0
        return reference in self.get_delivered_references()

    @staticmethod
    def get_reference(to_email_address, template_id, personalisation=None):
        """
        Method to return the standard reference given the variables the email is sent with.

        :param to_email_address: Emails recipient
        :param template_id: Emails template ID on Notify
        :param personalisation: Template parameters
        :return: Hashed string 'reference' to be passed to client.send_email_notification or self.send_email
        """
        personalisation_string = ",".join(
            map(str, personalisation.values())
        ) if personalisation else ""
        return hash_string(f"{to_email_address}|{template_id}|{personalisation_string}")

    def _log_email_error_message(self, email_obj: DMNotifyEmail, error):
        """Format a logical error message from the error response and send it to the logger"""

        if isinstance(error.message, str):
            error_messages = [error.message]
        else:
            error_messages = [
                f'{error.status_code} {error_message["error"]}: {error_message["message"]}'
                for error_message in error.message
            ]

        self._log(
            logging.ERROR,
            "Error sending email: {error_messages}",
            email_obj,
            extra={
                "error_messages": error_messages,
            },
        )

    def _log(self, lvl, msg, email_obj: DMNotifyEmail, *, extra: Optional[dict] = None, **kwargs):
        if extra is None:
            extra = {}
        extra.update({
            "client": self.__class__,
            "reference": email_obj.reference,
            "template_name_or_id": email_obj.template_name_or_id,
            "to_email_address": hash_string(email_obj.to_email_address),
        })
        self.logger.log(lvl, msg, extra)

    def send_email(
        self,
        to_email_address,
        template_name_or_id,
        personalisation=None,
        allow_resend=True,
        reference=None,
        reply_to_address_id=None,
        use_recent_cache=True
    ):
        """
        Method to send an email using the Notify api.

        :param to_email_address: String email address for recipient
        :param template_name_or_id: Template accessible on the Notify account,
                                    can either be a key to the `templates` dictionary or a Notify template ID.
        :param personalisation: The template variables, dict
        :param allow_resend: if False instantiate the delivered reference cache and ensure we are not sending duplicates
        :param reply_to_address_id: String id of reply-to email address. Must be set up in Notify config before use
        :param use_recent_cache: Use the client's cache of recently sent references. If set to False, any
                                 has_been_sent() calls will check the reference in the Notify API directly
        :return: response from the api. For more information see https://github.com/alphagov/notifications-python-client
        """
        template_id = self.templates.get(template_name_or_id, template_name_or_id)
        reference = reference or self.get_reference(to_email_address, template_id, personalisation)
        email_obj = DMNotifyEmail(to_email_address, template_name_or_id, reference, personalisation)

        if not allow_resend and self.has_been_sent(reference, use_recent_cache=use_recent_cache):
            self._log(
                logging.WARNING,
                "Email with reference '{reference}' has already been sent",
                email_obj,
            )
            return

        # NOTE how the potential replacement of the email address happens *after* the has_been_sent check and
        # reference generation
        final_email_address = (
            self._redirect_domains_to_address
            and self._redirect_domains_to_address.get(
                # splitting at rightmost @ should reliably give us the domain
                to_email_address.rsplit("@", 1)[-1].lower()
            )
        ) or to_email_address

        try:
            with log_external_request(service='Notify'):
                response = self.client.send_email_notification(
                    final_email_address,
                    template_id,
                    personalisation=personalisation,
                    reference=reference,
                    email_reply_to_id=reply_to_address_id
                )

        except HTTPError as e:
            self._log_email_error_message(email_obj, e)
            if isinstance(e.message, list) and \
                    any(msg["message"].startswith("Missing personalisation") for msg in e.message):
                raise EmailTemplateError(str(e))

            if isinstance(e.message, list) and \
                    any(msg["message"].startswith("email_address Not a valid email address") for msg in e.message):
                raise EmailInvalidError(str(e))

            raise EmailError(str(e))

        self._log(logging.INFO, f"Email with reference '{reference}' sent to Notify successfully", email_obj)

        self._update_cache(reference)

        return response
